/*
 * Volume control using an MCP4131 digital potentiometer connected via SPI
 *
 * Inputs are cap sense
 * Cap sense based on NG tx/rx cap sense sample
 *
 */

#include <avr/io.h>
#include <util/delay.h>

// General macros
#define output(directions,pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 102 // bit delay for 9600 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define settle_delay() _delay_us(100) // settle delay
#define char_delay() _delay_ms(10) // char delay
#define nloop 100 // loops to accumulate

// Pins for capacitive touch sense
// PA0 - tx
// PA1 ADC1 - button1
// PA2 ADC1 - button2
// PA3 ADC1 - button3
#define button_tx_port PORTA
#define button_tx_direction DDRA
#define button_tx_pin (1<<PA0)
#define button_rx1_mux ((0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0)) // ADC1
#define button_rx2_mux ((0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (0 << MUX0)) // ADC2
#define button_rx3_mux ((0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (1 << MUX0)) // ADC3


// Pins for spi communications
#define spi_port PORTA
#define spi_direction DDRA
#define spi_mosi_pin (1<<PA6)
#define spi_sck_pin (1<<PA4)
#define mcp4131_cs_port PORTB
#define mcp4131_cs_direction DDRB
#define mcp4131_cs_pin (1<<PB2)

// Pins for on board LED
#define ob_led_port PORTA
#define ob_led_direction DDRA
#define ob_led_pin (1<<PA7)


// write a byte out to the spi interface
// bit bang the lines as SPI hardware was difficult to setup
// SPI mode 0,0 - sck idles low, data clocked into mcp4131 on sck rising edge
void spi_send(uint8_t data)
{
    // sck idles in low state
    
    if(bit_test(data,7)) // start by sending most significant bit (msb)
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,6))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,5))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,4))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,3))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,2))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,1))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
    if(bit_test(data,0))
        set(spi_port, spi_mosi_pin);
    else
        clear(spi_port, spi_mosi_pin);
    set(spi_port, spi_sck_pin); // toggle the clock line
    _delay_us(1);
    clear(spi_port, spi_sck_pin);
    _delay_us(1);
    
}

// Set the volume via the digital pot
// level is 0 to 127
void set_volume(uint8_t level)
{
    // our digital pot is upside down so invert the volume
    // for instance if we write 0 we get full volume, if we write 127 we get mute
    level = 127 - level;
    
    clear(mcp4131_cs_port, mcp4131_cs_pin); // select the mcp4131 chip
    _delay_us(1); // wait a bit after chip select before sending any data
    spi_send(0); // send 'write wiper 0' command
    spi_send(level); // send actual volume level
    set(mcp4131_cs_port, mcp4131_cs_pin); // unselect the mcp4131 chip
}


int main(void)
{
    uint8_t count; // loop counter
    uint32_t up, down; // acculumations for capacitive touch sense
    
    uint8_t volume_up_pressed, volume_down_pressed, mute_pressed; // flags to say a button was pressed
    
    int volume = 0; // current volume setting
    
    
    //
    // set clock divider to /1
    //
    CLKPR = (1 << CLKPCE);
    CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
    
    //
    // initialize output pins
    //
    clear(ob_led_port, ob_led_pin);
    output(ob_led_direction, ob_led_pin);
    clear(button_tx_port, button_tx_pin);
    output(button_tx_direction, button_tx_pin);
    
    //
    // initialise the SPI interface
    //
    clear(spi_port, spi_mosi_pin);
    clear(spi_port, spi_sck_pin);
    output(spi_direction, spi_sck_pin);
    output(spi_direction, spi_mosi_pin);
    set(mcp4131_cs_port, mcp4131_cs_pin); // set mcp4131 pin high so that we don't select the chip yet
    output(mcp4131_cs_direction, mcp4131_cs_pin);
    
    //
    // init A/D for capacitive touch sense
    //
    ADMUX = (0 << REFS1) | (0 << REFS0) | button_rx3_mux ; // Vcc ref, button1 selected
    ADCSRA = (1 << ADEN) // enable
        | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128
    ADCSRB = (0 << ADLAR); // right adjust
    //DIDR0 = 0; // disable digital inputs on ADC pins we'll use?
    
    for(;;){
        
        // Reset the button states to not pressed before we read the buttons
        volume_up_pressed = 0;
        volume_down_pressed = 0;
        mute_pressed = 0;
        
        //
        // Read all of the buttons
        //
        
        // Read button 1
        ADMUX = (0 << REFS1) | (0 << REFS0) | button_rx1_mux ; // Vcc ref, button1 selected
        up = 0;
        down = 0;
        for(count = 0; count < nloop; count++)
        {
            //
            // settle, charge
            //
            settle_delay();
            set(button_tx_port, button_tx_pin);
            //
            // initiate conversion
            //
            ADCSRA |= (1 << ADSC);
            //
            // wait for completion
            //
            while (ADCSRA & (1 << ADSC))
                ;
            //
            // save result
            //
            up += ADC;
            //
            // settle, discharge
            //
            settle_delay();
            clear(button_tx_port, button_tx_pin);
            //
            // initiate conversion
            //
            ADCSRA |= (1 << ADSC);
            //
            // wait for completion
            //
            while (ADCSRA & (1 << ADSC))
                ;
            //
            // save result
            //
            down += ADC;
        }
        if((up - down) > 50000)
            volume_up_pressed = 1;
        
        // Read button 2
        ADMUX = (0 << REFS1) | (0 << REFS0) | button_rx2_mux ; // Vcc ref, button1 selected
        up = 0;
        down = 0;
        for(count = 0; count < nloop; count++)
        {
            //
            // settle, charge
            //
            settle_delay();
            set(button_tx_port, button_tx_pin);
            //
            // initiate conversion
            //
            ADCSRA |= (1 << ADSC);
            //
            // wait for completion
            //
            while (ADCSRA & (1 << ADSC))
                ;
            //
            // save result
            //
            up += ADC;
            //
            // settle, discharge
            //
            settle_delay();
            clear(button_tx_port, button_tx_pin);
            //
            // initiate conversion
            //
            ADCSRA |= (1 << ADSC);
            //
            // wait for completion
            //
            while (ADCSRA & (1 << ADSC))
                ;
            //
            // save result
            //
            down += ADC;
        }
        if((up - down) > 50000)
            mute_pressed = 1;
        
        // Read button 3
        ADMUX = (0 << REFS1) | (0 << REFS0) | button_rx3_mux ; // Vcc ref, button1 selected
        up = 0;
        down = 0;
        for(count = 0; count < nloop; count++)
        {
            //
            // settle, charge
            //
            settle_delay();
            set(button_tx_port, button_tx_pin);
            //
            // initiate conversion
            //
            ADCSRA |= (1 << ADSC);
            //
            // wait for completion
            //
            while (ADCSRA & (1 << ADSC))
                ;
            //
            // save result
            //
            up += ADC;
            //
            // settle, discharge
            //
            settle_delay();
            clear(button_tx_port, button_tx_pin);
            //
            // initiate conversion
            //
            ADCSRA |= (1 << ADSC);
            //
            // wait for completion
            //
            while (ADCSRA & (1 << ADSC))
                ;
            //
            // save result
            //
            down += ADC;
        }
        if((up - down) > 50000)
            volume_down_pressed = 1;
        
        // Show some button press feedback on the LED
        if(volume_up_pressed || volume_down_pressed || mute_pressed)
        {
            set(ob_led_port, ob_led_pin);
        }
        else
        {
            clear(ob_led_port, ob_led_pin);
        }
       
        //
        // Adjust volume based on the buttons
        //
        if(volume_up_pressed)
            volume = volume + 6;
        
        if(volume_down_pressed)
            volume = volume - 6;
        
        if(mute_pressed)
            volume = 0;

        // Volume must be in range 0 to 127, so range check it
        if( volume < 0)
            volume = 0;
        if(volume > 127)
            volume = 127;
        
        // Set the digital pot to the current volume
        set_volume((uint8_t)volume);
        
        // wait a bit before processing everything again
        _delay_ms(100);
   }
    return 0;
}
